"""
Run module.
"""
from __future__ import annotations

import typing
from collections import namedtuple

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._base.status import State
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.artifacts.crud import get_artifact_from_key
from digitalhub_core.entities.dataitems.crud import get_dataitem_from_key
from digitalhub_core.runtimes.builder import build_runtime
from digitalhub_core.utils.api import api_base_create, api_base_read, api_base_update, api_ctx_read
from digitalhub_core.utils.commons import ARTF, DTIT, FUNC, LOGS, RUNS, TASK
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.artifacts.entity import Artifact
    from digitalhub_core.entities.dataitems.entity import Dataitem
    from digitalhub_core.entities.runs.metadata import RunMetadata
    from digitalhub_core.entities.runs.spec import RunSpec
    from digitalhub_core.entities.runs.status import RunStatus
    from digitalhub_core.runtimes.base import Runtime


TaskString = namedtuple(
    "TaskString",
    ["function_kind", "task_kind", "function_name", "function_id", "task_id"],
)


class Run(Entity):
    """
    A class representing a run.
    """

    def __init__(
        self,
        uuid: str,
        kind: str,
        metadata: RunMetadata,
        spec: RunSpec,
        status: RunStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        uuid : str
            UUID.
        kind : str
            The type of the run.
        metadata : RunMetadata
            Metadata of the object.
        spec : RunSpec
            Specification of the object.
        status : RunStatus
            Status of the object.
        """
        super().__init__()
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        self.project = self.metadata.project
        self._obj_attr.extend(["project"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save run into backend.

        Parameters
        ----------
        uuid : str
            Ignore this parameter.

        Returns
        -------
        dict
            Mapping representation of Run from backend.
        """
        obj = self.to_dict(include_all_non_private=True)

        if not update:
            api = api_base_create(RUNS)
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_base_update(RUNS, self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        filename = (
            filename if filename is not None else f"run_{self.metadata.project}_{self.spec.task_id}_{self.id}.yaml"
        )
        write_yaml(filename, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.metadata.project)

    #############################
    #  Run Methods
    #############################

    def build(self) -> None:
        """
        Build run.

        Returns
        -------
        None
        """
        function = self._get_function()
        task = self._get_task()
        runtime = self._get_runtime()
        new_spec = runtime.build(function, task, self.to_dict())
        # Insert task string validation
        self.spec = build_spec(RUNS, self.kind, ignore_validation=True, **new_spec)
        self._set_status({"state": State.BUILT.value})
        self.save(update=True)

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        if self.spec.local_execution:
            if not self.status.state == State.BUILT.value:
                raise EntityError("Run is not in built state. Build it again.")
            self._set_status({"state": State.RUNNING.value})
            self.save(update=True)

        runtime = self._get_runtime()
        try:
            status = runtime.run(self.to_dict(include_all_non_private=True))
        except Exception as err:
            status = {"state": State.ERROR.value, "message": str(err)}
        self._set_status(status)
        self.save(update=True)
        return self

    def refresh(self) -> Run:
        """
        Get run from backend.

        Returns
        -------
        Run
            Run object.
        """
        api = api_base_read(RUNS, self.id)
        obj = self._context().read_object(api)
        refreshed_run = self.from_dict(RUNS, obj)
        self.kind = refreshed_run.kind
        self.metadata = refreshed_run.metadata
        self.spec = refreshed_run.spec
        self.status = refreshed_run.status
        return self

    def logs(self) -> dict:
        """
        Get run's logs from backend.
        Returns empty dictionary if context is local.

        Returns
        -------
        dict
            Logs from backend.
        """
        if self._context().local:
            return {}
        api = api_base_read(LOGS, self.id)
        return self._context().read_object(api)

    def _set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If status is not a dictionary or a Status object.
        """
        if not isinstance(status, dict):
            raise EntityError("Status must be a dictionary.")
        self.status = build_status(RUNS, **status)

    #############################
    #  Artifacts and Dataitems
    #############################

    def get_artifacts(self, output_key: str | None = None) -> Artifact | list[Artifact]:
        """
        Get artifact(s) from backend produced by the run through its key.

        Parameters
        ----------
        output_key : str, optional
            Key of the artifact to get. If not provided, returns all artifacts.

        Returns
        -------
        Artifact | list[Artifact]
            Artifact(s) from backend.
        """
        return self._get_objects(ARTF, get_artifact_from_key, output_key)

    def get_dataitem(self, output_key: str | None = None) -> Dataitem | list[Dataitem]:
        """
        Get dataitem(s) from backend produced by the run through its key.

        Parameters
        ----------
        output_key : str, optional
            Key of the dataitem to get. If not provided, returns all dataitems.

        Returns
        -------
        Dataitem | list[Dataitem]
            Dataitem(s) from backend.
        """
        return self._get_objects(DTIT, get_dataitem_from_key, output_key)

    def _get_objects(self, object_type: str, func: callable, output_key: str | None = None) -> list:
        """
        Get objects from backend produced by the run through its key.

        Parameters
        ----------
        object_type : str
            Type of the object to get.
        func : callable
            Function to get object from backend.
        output_key : str
            Key of the object to get. If not provided, returns all objects.

        Returns
        -------
        list
            Objects from backend.

        Raises
        ------
        EntityError
            If object type is not supported or if run has
            no result or if object with key is not found.
        """
        self.refresh()
        if object_type == DTIT:
            result = self.status.dataitems
        elif object_type == ARTF:
            result = self.status.artifacts
        else:
            raise EntityError(f"Object type '{object_type}' not supported.")
        if result is None:
            raise EntityError("Run has no result (maybe try when it finishes).")
        if output_key is not None:
            key = next((r.get("id") for r in result if r.get("key") == output_key), None)
            if key is None:
                raise EntityError(f"No {object_type} found with key '{output_key}'.")
            return func(key)
        return [func(r.get("id")) for r in result]

    #############################
    #  Functions and Tasks
    #############################

    def _parse_task_string(self) -> TaskString:
        """
        Parse task string.

        Returns
        -------
        None
        """
        kinds, func = self.spec.task.split("://")
        fnc_kind, tsk_kind = kinds.split("+")
        fnc_name, fnc_id = func.split("/")[1].split(":")
        return TaskString(fnc_kind, tsk_kind, fnc_name, fnc_id, self.spec.task_id)

    def _get_function(self) -> dict:
        """
        Get function from backend. Reimplemented to avoid circular imports.

        Returns
        -------
        dict
            Function from backend.
        """
        parsed = self._parse_task_string()
        api = api_ctx_read(self.metadata.project, FUNC, parsed.function_name, parsed.function_id)
        return self._context().read_object(api)

    def _get_task(self) -> dict:
        """
        Get task from backend. Reimplemented to avoid circular imports.

        Returns
        -------
        dict
            Task from backend.
        """
        parsed = self._parse_task_string()
        api = api_base_read(TASK, parsed.task_id)
        return self._context().read_object(api)

    #############################
    # Runtimes
    #############################

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        fnc_kind = self._parse_task_string().function_kind
        return build_runtime(fnc_kind)


def run_from_parameters(
    project: str,
    task: str,
    task_id: str,
    kind: str,
    uuid: str | None = None,
    inputs: dict | None = None,
    outputs: list | None = None,
    parameters: dict | None = None,
    local_execution: bool = False,
    **kwargs,
) -> Run:
    """
    Create run.

    Parameters
    ----------
    project : str
        Name of the project.
    task_id : str
        Identifier of the task associated with the run.
    task : str
        Name of the task associated with the run.
    kind : str
        The type of the run.
    uuid : str
        UUID.
    inputs : dict
        Inputs of the run.
    outputs : list
        Outputs of the run.
    parameters : dict
        Parameters of the run.
    local_execution : bool
        Flag to determine if object has local execution.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Run
        Run object.
    """
    uuid = build_uuid(uuid)
    metadata = build_metadata(
        RUNS,
        project=project,
        name=uuid,
    )
    spec = build_spec(
        RUNS,
        kind,
        task=task,
        task_id=task_id,
        inputs=inputs,
        outputs=outputs,
        parameters=parameters,
        local_execution=local_execution,
        **kwargs,
    )
    status = build_status(RUNS)
    return Run(
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def run_from_dict(obj: dict) -> Run:
    """
    Create run from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create run from.

    Returns
    -------
    Run
        Run object.
    """
    return Run.from_dict(RUNS, obj)



"""
Workflow operations module.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities.workflows.entity import workflow_from_dict, workflow_from_parameters
from digitalhub_core.utils.api import api_ctx_delete, api_ctx_read, api_ctx_update
from digitalhub_core.utils.commons import WKFL
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.workflows.entity import Workflow


def create_workflow(**kwargs) -> Workflow:
    """
    Create a new Workflow instance with the specified parameters.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Workflow
        An instance of the created workflow.
    """
    return workflow_from_parameters(**kwargs)


def create_workflow_from_dict(obj: dict) -> Workflow:
    """
    Create a new Workflow instance from a dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create the Workflow from.

    Returns
    -------
    Workflow
        Workflow object.
    """
    return workflow_from_dict(obj)


def new_workflow(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    embedded: bool = True,
    **kwargs,
) -> Workflow:
    """
    Create a new Workflow instance with the specified parameters.

    Parameters
    ----------
    project : str
        A string representing the project associated with this workflow.
    name : str
        The name of the workflow.
    kind : str
        Kind of the object.
    uuid : str
        UUID.
    description : str
        A description of the workflow.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Workflow
        An instance of the created workflow.
    """
    obj = create_workflow(
        project=project,
        name=name,
        description=description,
        kind=kind,
        embedded=embedded,
        uuid=uuid,
        **kwargs,
    )
    obj.save()
    return obj


def get_workflow(project: str, name: str, uuid: str | None = None) -> Workflow:
    """
    Retrieves workflow details from the backend.

    Parameters
    ----------

    project : str
        Name of the project.
    name : str
        The name of the workflow.
    uuid : str
        UUID.

    Returns
    -------
    Workflow
        Object instance.
    """
    api = api_ctx_read(project, WKFL, name, uuid=uuid)
    obj = get_context(project).read_object(api)
    return workflow_from_dict(obj)


def import_workflow(file: str) -> Workflow:
    """
    Import an Workflow object from a file using the specified file path.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Workflow
        Object instance.
    """
    obj = read_yaml(file)
    return workflow_from_dict(obj)


def delete_workflow(project: str, name: str, uuid: str | None = None) -> dict:
    """
    Delete workflow from the backend. If the uuid is not specified, delete all versions.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the workflow.
    uuid : str
        UUID.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_delete(project, WKFL, name, uuid=uuid)
    return get_context(project).delete_object(api)


def update_workflow(workflow: Workflow) -> dict:
    """
    Update a workflow.

    Parameters
    ----------
    workflow : Workflow
        The workflow to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(workflow.metadata.project, WKFL, workflow.metadata.name, uuid=workflow.id)
    return get_context(workflow.metadata.project).update_object(workflow.to_dict(), api)



"""
Remote store module.
"""
from __future__ import annotations

import typing

import requests
from digitalhub_core.stores.objects.base import Store, StoreConfig
from digitalhub_core.utils.file_utils import build_path

if typing.TYPE_CHECKING:
    import pandas as pd


class RemoteStoreConfig(StoreConfig):
    """
    Remote store configuration class.
    """


class RemoteStore(Store):
    """
    HTTP store class. It implements the Store interface and provides methods to fetch
    artifacts from remote HTTP based storage.
    """

    def __init__(self, name: str, store_type: str, config: RemoteStoreConfig) -> None:
        """
        Constructor.

        Parameters
        ----------
        config : RemoteStoreConfig
            Remote store configuration.

        See Also
        --------
        Store.__init__
        """
        super().__init__(name, store_type)
        self.config = config

    ############################
    # IO methods
    ############################

    def download(self, src: str, dst: str | None = None) -> str:
        """
        Method to download an artifact from the backend.

        Parameters
        ----------
        src : str
            The source location of the artifact.
        dst : str
            The destination of the artifact.

        Returns
        -------
        str
            The path of the downloaded artifact.
        """
        return self._registry.get(src, self.fetch_artifact(src, dst))

    def fetch_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to fetch an artifact from the remote storage and to register
        it on the paths registry.
        If the destination is not provided, a temporary folder will be created.

        Parameters
        ----------
        src : str
            The source location of the artifact.
        dst : str
            The destination of the artifact.

        Returns
        -------
        str
            Returns the path of the artifact.
        """
        dst = dst if dst is not None else self._build_temp(src)
        if not dst.endswith(".csv") or not dst.endswith(".parquet"):
            dst = build_path(dst, "temp.file")
        return self._download_file(src, dst)

    def upload(self, src: str, dst: str | None = None) -> str:
        """
        Method to upload an artifact to the backend. Please note that this method is not implemented
        since the remote store is not meant to upload artifacts.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Remote store does not support upload.")

    def persist_artifact(self, src: str, dst: str | None = None) -> str:
        """
        Method to persist an artifact. Note that this method is not implemented
        since the remote store is not meant to write artifacts.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Remote store does not support persist_artifact.")

    def write_df(self, df: pd.DataFrame, dst: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Note that this method is not implemented
        since the remote store is not meant to write dataframes.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Remote store does not support write_df.")

    ############################
    # Private helper methods
    ############################

    @staticmethod
    def _check_head(src) -> None:
        """
        Check if the source exists.

        Parameters
        ----------
        src : str
            The source location.

        Returns
        -------
        None

        Raises
        ------
        HTTPError
            If an error occurs while checking the source.
        """
        r = requests.head(src, timeout=60)
        r.raise_for_status()

    def _download_file(self, url: str, dst: str) -> str:
        """
        Method to download a file from a given url.

        Parameters
        ----------
        url : str
            The url of the file to download.
        dst : str
            The destination of the file.

        Returns
        -------
        str
            The path of the downloaded file.
        """
        self._check_head(url)
        self._check_local_dst(dst)
        with requests.get(url, stream=True, timeout=60) as r:
            r.raise_for_status()
            with open(dst, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return dst

    ############################
    # Store interface methods
    ############################

    @staticmethod
    def is_local() -> bool:
        """
        Check if the store is local.

        Returns
        -------
        bool
            False
        """
        return False



"""
Client builder module.
"""
from __future__ import annotations

import typing

from digitalhub_core.client.objects.dhcore import ClientDHCore
from digitalhub_core.client.objects.local import ClientLocal

if typing.TYPE_CHECKING:
    from digitalhub_core.client.objects.base import Client


class ClientBuilder:
    """
    The client builder. It implements the builder pattern to create a client instance.
    """

    def __init__(self) -> None:
        self._client = None

    def build(self, local: bool = False) -> Client:
        """
        Method to create a client instance.

        Parameters
        ----------
        local : bool
            Whether to create a local client or not.

        Returns
        -------
        Client
            Returns the client instance.
        """
        if self._client is None:
            if local:
                self._client = ClientLocal()
            else:
                self._client = ClientDHCore()
        return self._client


def get_client(local: bool = False) -> Client:
    """
    Wrapper around ClientBuilder.build.

    Parameters
    ----------
    local : bool
        Whether to create a local client or not.

    Returns
    -------
    Client
        The client instance.
    """
    return client_builder.build(local)


client_builder = ClientBuilder()



"""
Model object module.
"""
from __future__ import annotations


class ModelObj:
    """
    Base class for all entities.
    """

    def to_dict(self) -> dict:
        """
        Return object as dict with all non private keys.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        obj = {}
        for key, value in self.__dict__.items():
            if not key.startswith("_") and value is not None:
                if hasattr(value, "to_dict"):
                    sub_dict = value.to_dict()
                    if sub_dict:
                        obj[key] = sub_dict
                else:
                    obj[key] = value
        return obj

    def _any_setter(self, **kwargs) -> None:
        """
        Set any attribute of the object.

        Parameters
        ----------
        kwargs : dict
            A dictionary containing the attributes of the entity instance.

        Returns
        -------
        None
        """
        for k, v in kwargs.items():
            if k not in self.__dict__:
                setattr(self, k, v)

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict())



"""
Common filesystem utils.
"""
from __future__ import annotations

import shutil
from pathlib import Path

####################
# Paths utils
####################


def build_path(*args) -> str:
    """
    Build path from args list.

    Parameters
    ----------
    *args
        Arguments list.

    Returns
    -------
    str
        The path.
    """
    return str(Path(*args))


def get_absolute_path(*args) -> str:
    """
    Return absolute path.

    Parameters
    ----------
    *args
        Arguments list.

    Returns
    -------
    str
        The absolute path.
    """
    return str(Path(*args).absolute())


def check_path(path: str) -> bool:
    """
    Check if the path exists.

    Parameters
    ----------
    path : str
        The path to check.

    Returns
    -------
    bool
        True if the path exists, False otherwise.
    """
    try:
        return Path(path).exists()
    except OSError:
        return False


def clean_all(path: str) -> None:
    """
    Remove dir and all it's contents.

    Parameters
    ----------
    path : str
        The directory path.

    Returns
    -------
    None

    Raises
    ------
    Exception
        If the directory cannot be removed.
    """
    shutil.rmtree(path)


####################
# Directories utils
####################


def get_dir(path: str) -> str:
    """
    Return directory path.

    Parameters
    ----------
    path : str
        The path.

    Returns
    -------
    str
        The directory path.
    """
    pth = Path(path)
    if pth.is_file():
        return str(pth.parent)
    return str(pth)


def check_dir(path: str) -> bool:
    """
    Check if path is a directory.

    Parameters
    ----------
    path : str
        The path to check.

    Returns
    -------
    bool
        True if the directory exists, False otherwise.
    """
    try:
        return Path(path).is_dir()
    except OSError:
        return False


def make_dir(*args) -> None:
    """
    Create a directory.

    Parameters
    ----------
    *args
        Arguments list.

    Returns
    -------
    None
    """
    pth = Path(*args).absolute()
    if pth.suffix != "":
        pth = pth.parent
    pth.mkdir(parents=True, exist_ok=True)


####################
# Files utils
####################


def check_file(path: str) -> bool:
    """
    Check if a path is a file.

    Parameters
    ----------
    path : str
        The file path.

    Returns
    -------
    bool
        True if the file exists, False otherwise.
    """
    try:
        return Path(path).is_file()
    except OSError:
        return False


def copy_file(src: str, dst: str) -> str:
    """
    Copy local file to destination.

    Parameters
    ----------
    src : str
        The source file.
    dst : str
        The destination file/directory.

    Returns
    -------
    str
        The copied file path.
    """
    return shutil.copy(src, dst)


def is_python_module(src: str) -> bool:
    """
    Check if a file is a python module.

    Parameters
    ----------
    src : str
        The file path.

    Returns
    -------
    bool
        True if the file is a python module, False otherwise.
    """
    return Path(src).suffix == ".py" and check_file(src)



"""
DHCore Client module.
"""
from __future__ import annotations

import os

import requests
from digitalhub_core.client.objects.base import Client
from digitalhub_core.utils.exceptions import BackendError


class ClientDHCore(Client):
    """
    The client. It's a singleton. Use the builder to get an instance.
    It is used to make requests to the DHCore API.
    """

    def create_object(self, obj: dict, api: str) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        return self._call("POST", api, json=obj)

    def read_object(self, api: str) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict
            The object.
        """
        return self._call("GET", api)

    def update_object(self, obj: dict, api: str) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        return self._call("PUT", api, json=obj)

    def delete_object(self, api: str) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            A generic dictionary.
        """
        resp = self._call("DELETE", api)
        if isinstance(resp, bool):
            resp = {"deleted": resp}
        return resp

    def _call(self, call_type: str, api: str, **kwargs) -> dict:
        """
        Make a call to the DHCore API.
        Keyword arguments are passed to the session.request function.

        Parameters
        ----------
        call_type : str
            The type of call to make.
        api : str
            The api to call.
        **kwargs
            Keyword arguments.

        Returns
        -------
        dict
            The response object.
        """
        url = self._get_endpoint() + api
        response = None
        try:
            response = requests.request(call_type, url, timeout=60, **kwargs)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            if response is None:
                msg = "Unable to connect to DHCore backend."
            else:
                msg = f"Backend error: {response.status_code} - {response.text}"
            raise BackendError(msg)

    @staticmethod
    def _get_endpoint() -> str:
        """
        Get DHub Core endpoint environment variables.

        Returns
        -------
        str
            DHub Core endpoint environment variables.

        Raises
        ------
        Exception
            If the endpoint of DHCore is not set in the env variables.
        """
        endpoint = os.getenv("DHUB_CORE_ENDPOINT")
        if endpoint is None:
            raise BackendError("Endpoint not set as environment variables.")
        if endpoint.endswith("/"):
            endpoint = endpoint[:-1]
        return endpoint

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            False
        """
        return False



"""
Local Client module.
"""
from digitalhub_core.client.objects.base import Client
from digitalhub_core.utils.commons import ARTF, DTIT, FUNC, MDLS, PROJ, RUNS, TASK, WKFL
from digitalhub_core.utils.exceptions import BackendError


class ClientLocal(Client):
    """
    The local client. Use the builder to get an instance.
    It is used to keep objects in memory.
    """

    def __init__(self) -> None:
        super().__init__()
        self._db = {
            PROJ: {},
            FUNC: {},
            WKFL: {},
            RUNS: {},
            TASK: {},
            ARTF: {},
            DTIT: {},
            MDLS: {},
        }

    def create_object(self, obj: dict, api: str) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        parsed = self._parse_api(api)

        # Project, Task, Run
        if len(parsed) == 1:
            (dto,) = parsed
            name = obj.get("metadata", {}).get("name")
            self._db[dto][name] = obj

        # Artifact, Dataitem, Model, Function, Workflow
        if len(parsed) == 2:
            project, dto = parsed
            name = obj.get("metadata", {}).get("name")
            uuid = obj.get("metadata", {}).get("version")
            self._db[dto].setdefault(project, {}).setdefault(name, {})
            self._db[dto][project][name][uuid] = obj
            self._db[dto][project][name]["latest"] = obj

        return obj

    def read_object(self, api: str) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict or None
            The object, or None if it doesn't exist.
        """
        parsed = self._parse_api(api)

        # Project, Task, Run
        if len(parsed) == 2:
            dto, name = parsed
            obj = self._db.get(dto, {}).get(name)
            if dto == PROJ:
                obj = self._get_project_spec(obj, name)
            msg = f"Object '{dto}' named '{name}' not found"

        # Artifact, Dataitem, Model, Function, Workflow
        elif len(parsed) == 4:
            project, dto, name, uuid = parsed
            obj = self._db.get(dto, {}).get(project, {}).get(name, {}).get(uuid)
            msg = f"Object '{dto}' named '{name}:{uuid}' for project '{project}' not found"

        if obj is None or not isinstance(obj, dict):
            raise BackendError(msg)

        return obj

    def update_object(self, obj: dict, api: str) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        parsed = self._parse_api(api)
        msg = None

        try:
            # Project, Task, Run
            if len(parsed) == 2:
                dto, name = parsed
                self._db[dto][name] = obj
                msg = f"Object '{dto}' named '{name}' not found"

            # Artifact, Dataitem, Model, Function, Workflow
            elif len(parsed) == 4:
                project, dto, name, uuid = parsed
                self._db[dto][project][name][uuid] = obj
                msg = f"Object '{dto}' named '{name}:{uuid}' for project '{project}' not found"

        except KeyError:
            raise BackendError(msg)

        return obj

    def delete_object(self, api: str) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            A generic dictionary.
        """
        parsed = self._parse_api(api)
        fallback = "No element found"

        # Project, Task, Run
        if len(parsed) == 2:
            dto, name = parsed
            deleted = self._db[dto].pop(name, fallback)

        # Artifact, Dataitem, Model, Function, Workflow
        # No uuid
        elif len(parsed) == 3:
            project, dto, name = parsed
            deleted = self._db[dto][project].pop(name, fallback)

        # uuid
        elif len(parsed) == 4:
            project, dto, name, uuid = parsed
            deleted = self._db[dto][project][name].pop(uuid, fallback)

        return {"deleted": deleted}

    @staticmethod
    def _parse_api(api: str) -> list[str]:
        """
        Parse the api to get information about the object.

        Parameters
        ----------
        api : str
            The api to parse.

        Returns
        -------
        list[str]
            The parsed api.
        """
        api = api.removeprefix("/api/v1/")
        splitted = api.split("/")
        if splitted[0] == "-":
            return splitted[1:]
        return splitted

    def _get_project_spec(self, obj: dict, name: str) -> dict:
        """
        Read the project spec.

        Parameters
        ----------
        obj : dict
            The project object.
        name : str
            The project name.

        Returns
        -------
        dict
            The project object with the spec.
        """
        if obj is None or not isinstance(obj, dict):
            raise BackendError(f"Project not found: {name}")
        for i in [FUNC, WKFL, ARTF, DTIT]:
            objs = self._db.get(i, {}).get(name, {})
            obj["spec"][i] = []
            for _, j in objs.items():
                for k, v in j.items():
                    if k != "latest":
                        obj["spec"][i].append(v)
        return obj

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            True
        """
        return True



"""
Common constants used in the SDK.
"""

####################
# Entities
####################

PROJ = "projects"
FUNC = "functions"
WKFL = "workflows"
TASK = "tasks"
RUNS = "runs"
LOGS = "logs"
ARTF = "artifacts"
DTIT = "dataitems"
MDLS = "models"



"""
Dataitem specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec
from pydantic import BaseModel


class DataitemSpec(Spec):
    """
    Dataitem specifications.
    """

    def __init__(self, key: str | None = None, path: str | None = None, **kwargs) -> None:
        """
        Constructor.

        Parameters
        ----------
        key : str
            The key of the dataitem.
        path : str
            The path of the dataitem.
        **kwargs
            Keyword arguments.
        """
        self.key = key
        self.path = path

        self._any_setter(**kwargs)


class DataitemParams(BaseModel):
    """
    Dataitem parameters.
    """

    key: str = None
    """The key of the dataitem."""
    path: str = None
    "The path of the dataitem."
